/**
 * EasyBeans
 * Copyright (C) 2006-2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: DefaultAnnotationProcessor.java 4700 2009-02-24 15:05:59Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.annotation.processor;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

/**
 * The {@link DefaultAnnotationProcessor} process an Object and call
 * {@link IAnnotationHandler}'s methods for each found annotation.
 * @author Guillaume Sauthier
 */
public class DefaultAnnotationProcessor implements IAnnotationProcessor {

	/**
	 * Registered {@link IAnnotationHandler}.
	 */
	private List<IAnnotationHandler> handlers;

	/**
	 * Construct a default {@link DefaultAnnotationProcessor}.
	 */
	public DefaultAnnotationProcessor() {
		handlers = new ArrayList<IAnnotationHandler>();
	}

	/**
	 * Process the given object.
	 * @param target processed object
	 * @throws ProcessorException
	 */
	public void process(final Object target) throws ProcessorException {
		// Extract the Object's Class and process it
		Class<?> clazz = target.getClass();
		process(target, clazz);
	}

	/**
	 * Process the given class and uses the given target
	 * instance for injection if any.
	 * @param target Object for injection
	 * @param clazz Class to be processed
	 * @throws ProcessorException
	 */
	public void process(final Object target, final Class<?> clazz) throws ProcessorException {

		// First process parent
		if (clazz.getSuperclass() != null) {
			process(target, clazz.getSuperclass());
		}

		// Process class level annotations
		processClassAnnotations(target, clazz);

		// Process method level annotations
		Method[] methods = clazz.getDeclaredMethods();
		for (Method method : methods) {
            processMethodAnnotations(target, method);
		}

		// Process field level annotations
		Field[] fields = clazz.getDeclaredFields();
		for (Field field : fields) {
			processFieldAnnotations(target, field);
		}
	}

	/**
	 * Process a field and its associated annotations
	 * @param target instance
	 * @param field processed field
	 * @throws ProcessorException
	 */
	protected void processFieldAnnotations(final Object target, final Field field) throws ProcessorException {

		// Extract annotations
		Annotation[] annotations = field.getAnnotations();
		if (annotations != null) {
			for (Annotation annotation : annotations) {
				for (IAnnotationHandler handler : handlers) {

					// If a suitable handler is found
					if (handler.isSupported(annotation.annotationType())) {
						handler.process(annotation, field, target);
					}
				}
			}
		}
	}

	/**
	 * Process a method and its associated annotations
	 * @param target instance
	 * @param method processed method
	 * @throws ProcessorException
	 */
	protected void processMethodAnnotations(final Object target, final Method method) throws ProcessorException {

		// Extract annotations
		Annotation[] annotations = method.getAnnotations();
		if (annotations != null) {
			for (Annotation annotation : annotations) {
				for (IAnnotationHandler handler : handlers) {

					// If a suitable handler is found
					if (handler.isSupported(annotation.annotationType())) {
						handler.process(annotation, method, target);
					}
				}
			}
		}
	}

	/**
	 * Process a class and its associated annotations
	 * @param target instance
	 * @param clazz processed class
	 * @throws ProcessorException
	 */
	protected void processClassAnnotations(final Object target, final Class<?> clazz) throws ProcessorException {

		// Extract annotations
		Annotation[] annotations = clazz.getAnnotations();
		if (annotations != null) {
			for (Annotation annotation : annotations) {
				for (IAnnotationHandler handler : handlers) {

					// If a suitable handler is found
					if (handler.isSupported(annotation.annotationType())) {
						handler.process(annotation, clazz, target);
					}
				}
			}
		}
	}

	/**
	 * Register the given handler
	 * @param handler registered handler
     */
	public void addAnnotationHandler(final IAnnotationHandler handler) {
        handlers.add(handler);
	}

	/**
	 * Unregister the given handler
	 * @param handler unregistered handler
	 */
	public void removeAnnotationHandler(final IAnnotationHandler handler) {
        handlers.remove(handler);
	}

}
